/******************************************************************************
 *                                                                            *
 * LISTSERV V2 - Standalone incoming mail interface (unix)                    *
 *                                                                            *
 * Copyright L-Soft international 1994,2000,2020, 2024 - All rights reserved  *
 *                                                                            *
 * Syntax: add entries to /etc/aliases using either of the following forms    *
 *                                                                            *
 *         listserv: "|/.../lsv_amin spool_dir listserv"                      *
 *         listserv: "|/.../lsv_amin -t listserv"                             *
 *                                                                            *
 * Where "..." represents the path to the executable (which should run with   *
 * setuid listserv rather than root, for increased security), "spool_dir" is  *
 * the absolute path to the LISTSERV spool directory, and the third parameter *
 * is a copy of the userid the entry is for (the name before the colon). If   *
 * no spool directory is specified, the default (defined in the Makefile and  *
 * also in lsv_amin.h) will be used.                                          *
 *                                                                            *
 * You must define entries for the reserved addresses LISTSERV and            *
 * owner-LISTSERV. In addition, for each list you must define entries for     *
 * listname, owner-listname, listname-request and listname-server. Example:   *
 *                                                                            *
 *  listserv: "|/usr/listserv/lsv_amin -t listserv"                           *
 *  owner-listserv: "|/usr/listserv/lsv_amin -t owner-listserv"               *
 *                                                                            *
 *  test-l: "|/usr/listserv/lsv_amin -t test-l"                               *
 *  test-l-request: "|/usr/listserv/lsv_amin -t test-l-request"               *
 *  test-l-server: "|/usr/listserv/lsv_amin -t test-l-server"                 *
 *  owner-test-l: "|/usr/listserv/lsv_amin -t owner-test-l"                   *
 *                                                                            *
 *****************************************************************************/


#include <sys/types.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <time.h>
#include <sys/stat.h>
#include <syslog.h>
#include <signal.h>

#include <unistd.h>
#include <ctype.h>

#include "lsv_amin.h"

#define Err_Invalid 200
#define Err_BadSyntax 201
#define Err_SysCallFailed 202
#define Err_NoFilename 203


#define TIME_LEN 8
#define TIME_CONV "%.8x"
#define PROC_LEN 5
#define PROC_CONV "%.6x"
#define MAXTRIES 10
#define JOB_EXT ".job"
#define JOB_PRE "in"
#define LISTSERV "LISTSERV"
#define PROGNAME "lsv_amin"
#define PROGVER "1.0d"


static unsigned char L_TA2E[] = {
 /* 0 */ 0x00,0x01,0x02,0x03,0x37,0x2D,0x2E,0x2F,
         0x16,0x05,0x25,0x0B,0x0C,0x0D,0x0E,0x0F,
 /* 1 */ 0x10,0x11,0x12,0x13,0x3C,0x3D,0x32,0x26,
         0x18,0x19,0x3F,0x27,0x1C,0x1D,0x1E,0x1F,
 /* 2 */ 0x40,0x5A,0x7F,0x7B,0x5B,0x6C,0x50,0x7D,
         0x4D,0x5D,0x5C,0x4E,0x6B,0x60,0x4B,0x61,
 /* 3 */ 0xF0,0xF1,0xF2,0xF3,0xF4,0xF5,0xF6,0xF7,
         0xF8,0xF9,0x7A,0x5E,0x4C,0x7E,0x6E,0x6F,
 /* 4 */ 0x7C,0xC1,0xC2,0xC3,0xC4,0xC5,0xC6,0xC7,
         0xC8,0xC9,0xD1,0xD2,0xD3,0xD4,0xD5,0xD6,
 /* 5 */ 0xD7,0xD8,0xD9,0xE2,0xE3,0xE4,0xE5,0xE6,
         0xE7,0xE8,0xE9,0xAD,0xE0,0xBD,0x5F,0x6D,
 /* 6 */ 0x79,0x81,0x82,0x83,0x84,0x85,0x86,0x87,
         0x88,0x89,0x91,0x92,0x93,0x94,0x95,0x96,
 /* 7 */ 0x97,0x98,0x99,0xA2,0xA3,0xA4,0xA5,0xA6,
         0xA7,0xA8,0xA9,0xC0,0x4F,0xD0,0xA1,0x07,
 /* 8 */ 0x20,0x21,0x22,0x23,0x24,0x2A,0x06,0x17,
         0x28,0x29,0x15,0x2B,0x2C,0x09,0x0A,0x1B,
 /* 9 */ 0x30,0x31,0x1A,0x33,0x34,0x35,0x36,0x08,
         0x38,0x39,0x3A,0x3B,0x04,0x14,0x3E,0xE1,
 /* A */ 0x41,0x42,0x43,0x44,0x45,0x46,0x47,0x48,
         0x49,0x51,0x52,0x53,0x54,0x55,0x56,0x57,
 /* B */ 0x58,0x59,0x62,0x63,0x64,0x65,0x66,0x67,
         0x68,0x69,0x70,0x71,0x72,0x73,0x74,0x75,
 /* C */ 0x76,0x77,0x78,0x80,0x8A,0x8B,0x8C,0x8D,
         0x8E,0x8F,0x90,0x9A,0x9B,0x9C,0x9D,0x9E,
 /* D */ 0x9F,0xA0,0xAA,0xAB,0xAC,0xBB,0xAE,0xAF,
         0xB0,0xB1,0xB2,0xB3,0xB4,0xB5,0xB6,0xB7,
 /* E */ 0xB8,0xB9,0xBA,0x4A,0xBC,0xFC,0xBE,0xBF,
         0xCA,0xCB,0xCC,0xCD,0xCE,0xCF,0xDA,0xDB,
 /* F */ 0xDC,0xDD,0xDE,0xDF,0xEA,0xEB,0xEC,0xED,
         0xEE,0xEF,0xFA,0xFB,0x6A,0xFD,0xFE,0xFF
};


struct err_entry {
  int msg_num;
  char *msg_text;
};


struct err_entry err_repos[] =
{  {  Err_Invalid,
      "Program error.*"
   },


   {  Err_BadSyntax,
#ifdef MAIL_PP
      "Syntax is, lsv_amin spool-directory to-user from-user"
#else
      "Syntax is, lsv_amin spool-directory to-user"
#endif
   },


   {  Err_SysCallFailed,
      "An unknown system call failed.*"
   },


   {  Err_NoFilename,
      "The incoming mail directory is full."
   }
};


char *user = "", *routine_in_err = "";


/* - FILE_EXISTS -
 * Return 1/0 indicating whether or not the indicated file exists.
 */
int file_exists( path)
char *path;
{
  struct stat stat_buff;


  return( !lstat( path, &stat_buff));
}

/* - ERR_EXIT -
 * Construct an appropriate error message and log it to stderr and syslog.
 */
int err_exit( rc)
int rc;
{
  long last_entry, wh = 0;
  char *msg = "";


  if( *user != '\0')
  {
    syslog( LOG_ERR, "Unable to deliver mail to: %s", user);
    fprintf( stderr, "lsv_amin: Unable to deliver mail to: %s\n", user);
  }


  if( rc == Err_SysCallFailed)
  {
    msg = (char *)malloc( 80);
    if( *routine_in_err == '\0') msg = "An unknown system call failed.*";
    else sprintf( msg, "A call to %s() failed.*", routine_in_err);
  }
  else
  {
    last_entry = (long)err_repos + sizeof( err_repos);


    for( ; (long)&err_repos[ wh] < last_entry && *msg == '\0'; wh++)
      if( rc == err_repos[ wh].msg_num) msg = err_repos[ wh].msg_text;


    if( *msg == '\0') msg = err_repos[ 0].msg_text;
  }


  if( msg[ strlen( msg) - 1] == '*')
  {
    msg[ strlen( msg) - 1] = '\0';
    syslog( LOG_ERR, "**Error(%d)** %s", errno, msg);
    fprintf( stderr, "lsv_amin: **Error(%d)** %s\n", errno, msg);
  }
  else
  {
    syslog( LOG_ERR, "**Error** %s", msg);
    fprintf( stderr, "lsv_amin: **Error** %s\n", msg);
  }


  closelog();
  exit( rc);
}

/* - SYSCALL_FAILED -
 * Save the name of the system call that bombed and log the error.
 */
int syscall_failed( which)
char *which;
{
  routine_in_err = which;
  err_exit( Err_SysCallFailed);


  exit( Err_SysCallFailed);  /* Shouldn't get here! */
}

/* - PUTSTR_EBC -
 * Convert '\0' terminated string of ASCII characters to EBCDIC and
 * write them to a file.
 */
void putstr_EBC( st, file)
unsigned char *st;
FILE *file;
{
  unsigned char *ch;
  int l;


#ifdef DBG_ASCII
  fprintf( file, "%s\n", st);
  fprintf( stdout, "%s\n", st);
#else
  l = strlen(st);
  putc(l >> 8, file);
  putc(l & 255, file);


  for( ch = st; *ch; ch++)
    if( fputc( L_TA2E[ *ch], file) == EOF) syscall_failed( "fputc");
#endif


  return;
}

/* - LSV_AMIN -
 */
int main( narg, args)
int narg;
char **args;
{
  long staging, proc;
  int ii, ntry, plen, xam;
  char *conv, *fsuffix, *final, *workfile, *fprefix, *w, *r, Fromline[1024],
    *fromid = "", uid[256], *spool_dir = "", **which, *def_spool, *callname;
  time_t now;
  FILE *out = 0, *pidfile;
  unsigned char buf[ 65537];


  fsuffix = (char *) malloc( sizeof JOB_EXT);
  strcpy( fsuffix, JOB_EXT);
  fprefix = (char *) malloc( sizeof JOB_PRE);
  strcpy( fprefix, JOB_PRE);
  def_spool = (char *) malloc( sizeof SPOOL_DIR);
  strcpy( def_spool, SPOOL_DIR);


  openlog( "LSV:lsv_amin", LOG_PID, LOG_FACILITY);


  callname = args[ 0];
  for( w = callname; *w; w++) if( *w == '/') callname = w + 1;
  if( strcmp( callname, PROGNAME)) user = callname;


  if( narg > 1)
  {
    if( !strcmp( args[ 1], "-v"))
    {
      fprintf( stdout, "%s release %s\n", PROGNAME, PROGVER);
      exit( 0);
    }


    if( *args[ 1] == '-') which = 0;
    else which = &spool_dir;
  }


  for( ii= 1; ii < narg; ii++)
  {
    w = args[ ii];
    if( which)
    {
      *which = w;
      which = 0;
    }
#ifdef MAIL_PP
    else if( !strcmp( w, "-f")) which = &fromid;
#endif
    else if( !strcmp( w, "-s")) which = &spool_dir;
    else if( !strcmp( w, "-t")) which = &user;
    else
    {
      if( *spool_dir == '\0') spool_dir = w;
      else if( *user == '\0') user = w;
#ifdef MAIL_PP
      else if( *fromid == '\0') fromid = w;
#endif
      else (void) err_exit( Err_BadSyntax);
      which = 0;
    }
  }


  if( !*spool_dir) spool_dir = def_spool;
#ifdef MAIL_PP
  if( !*fromid || !*user) (void) err_exit( Err_BadSyntax);
#else
  if( !*user) (void) err_exit( Err_BadSyntax);
#endif


#ifdef DBG_ASCII
  fprintf( stdout, "To<%s> Sp<%s>\n", user, spool_dir);
#endif


  for (r = user, w = uid; *r; *w++ = toupper(*r), r++);
  *w = '\0';


  workfile = (char *)malloc( strlen( spool_dir) + TIME_LEN + PROC_LEN +
    strlen( fprefix) + 3);
  final = (char *)malloc( strlen( spool_dir) + TIME_LEN + PROC_LEN +
    strlen(fprefix) + strlen( fsuffix) + 3);
  conv = (char *)malloc( (sizeof staging) * 2 + 1);


  strcpy( final, spool_dir);
  if( final[ (ii = strlen( final)) - 1] != '/')
  {
    final[ ii] = '/';
    final[ ii + 1] = '\0';
  }
  plen = strlen( final);


  for( ntry = 1; ntry <= MAXTRIES && !out; ntry++)
  {
    final[ plen] = '\0';
    strcat( final, fprefix);


    time( &now);
    sprintf( conv, TIME_CONV, (staging = now));
    strcat( final, (conv + strlen( conv)) - TIME_LEN);


    sprintf( conv, PROC_CONV, getpid());
    strcat( final, (conv + strlen( conv)) - (PROC_LEN + 1));
    final[ strlen( final) - (PROC_LEN + 1)] = '-';


    strcpy( workfile, final);
    strcat( final, fsuffix);


    if( !file_exists( workfile) && !file_exists( final))
    {
      if( !(out = fopen( workfile, "w"))) syscall_failed( "fopen");
    }
    else sleep( 1);
  }


  if( out)
  {
    if( fwrite("*LSV-V*", 7, 1, out) <= 0) syscall_failed( "fwrite");


    /* See if it is an administrative (owner-xxx/xxx-request) message */
    w = uid + strlen(uid);
    xam = !strncmp(uid, "OWNER-", 6) ||
          (w - uid > 8 && !strcmp(w - 8, "-REQUEST")) ||
          (w - uid > 7 && !strcmp(w - 7, "-SERVER"));
    /* Format and write the header */
    sprintf(buf, "1 M %s %s *MAILER* INCOMING MAIL",
                 uid,
                 (xam || !strcmp(uid, LISTSERV)) ? "STANDARD" :
                                                   "$NOJOB$");


    ii = strlen(buf);
#ifndef DBG_ASCII
    putc(ii >> 8, out);
    putc(ii & 255, out);
    fwrite(buf, ii, 1, out);
    putc(0, out);
    putc(0, out);
#else
    fwrite(buf, ii, 1, out);
    fputc( '\n', out);
    fputc( '\n', stdout);
#endif


    /* For administrative messages, generate X-ADMMAIL JCL */
#ifndef MAIL_PP
    fgets(Fromline, sizeof(Fromline), stdin);
#endif
    if (xam) {
        putstr_EBC("// JOB Echo=No", out);
#ifndef MAIL_PP
        for (fromid = Fromline; *fromid && *fromid++ != ' ';);
#endif
        for (w = fromid; *w && *w != ' ' && *w != '\n'; w++);
        *w = '\0';
        if( *fromid == '<' && fromid[ strlen( fromid) - 1] == '>')
        {
          fromid[ strlen( fromid) - 1] = '\0';
          fromid++;
        }
        sprintf(buf, "X-ADMMAIL %s %s", user, fromid);
        putstr_EBC(buf, out);
        putstr_EBC("//DATA DD *,EOF", out);
    }


    for (; fgets(buf, sizeof(buf), stdin);) {
        if (buf[0])
                buf[strlen(buf) - 1] = '\0';
        putstr_EBC(buf, out);
    }


    if( fclose( out)) syscall_failed( "fclose");
    if( link( workfile, final)) syscall_failed( "link");
    if( unlink( workfile)) syscall_failed( "unlink");


    syslog( LOG_INFO, "Mail for %s (from %d/%d) delivered to LISTSERV.",
      user, getuid(), geteuid());


        sprintf(buf, "%s/listserv.PID", spool_dir);
        if (pidfile = fopen(buf, "r")) {
                fscanf(pidfile, "%lu\n", &proc);
                kill(proc, SIGUSR1);
                fclose(pidfile);
        }
  }
  else err_exit( Err_NoFilename);


  exit( 0);
}

